import logging
from logging.handlers import RotatingFileHandler
import json
from .default_logger import DefaultLogger

class LoggerSingleton:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(LoggerSingleton, cls).__new__(cls)
            cls._instance._configure_logger()
        return cls._instance

    def _configure_logger(self):
        self.logger = logging.getLogger("custom_logger")
        self.logger.setLevel(logging.INFO)

        file_handler = RotatingFileHandler("app.log", maxBytes=5*1024*1024, backupCount=5)
        file_format = logging.Formatter('%(message)s')
        file_handler.setFormatter(file_format)

        console_handler = logging.StreamHandler()
        console_format = logging.Formatter('%(message)s')
        console_handler.setFormatter(console_format)

        self.logger.addHandler(file_handler)
        self.logger.addHandler(console_handler)

    def get_logger(self):
        return self.logger

    def log_transaction(self, log_entry: DefaultLogger):
        log_method = getattr(self.logger, log_entry.level.lower(), self.logger.info)
        log_method(log_entry.json())


from pydantic_settings import BaseSettings
from typing import Callable


class Settings(BaseSettings):
    # MONGO
    MONGO_URI: str = "mongodb://localhost:27017"
    MONGO_DB_NAME: str = "default_db"
    MONGO_TIMEOUT_MS: int = 500
    MONGO_MAX_POOL_SIZE: int = 20
    # APP
    APP_NAME: str = "default_app"
    SERVICE_NAME: str = "default_service"
    NAMESPACE: str = "default_namespace"
    API_VERSION: str = "v1"
    ENABLE_DOCS: bool = False
    # LOGS
    VERSION_LOG: str = "1.0.0"
    #SESSION
    IGNORE_SESSION: bool = False
    #EXTRA SETTINGS
    ENV: str = "DEV"
    HTTP_PORT: int = 8080
    RETRY: int = 3
    TIME_CLOSE_CIRCUIT: int = 5
    MAX_ATTEMPTS_OPEN_CIRCUIT: int = 10
    MAX_TRYING_TRANSACTIONS: int = 3
    URL_BASE: str = "http//localhost:8080/puntodeventa/api/v1"

    class Config:
        env_file = ".env"
        @classmethod
        def parse_env_var(cls, field: str, value: str) -> list:
            return value.split(',')


def _configure_initial_settings() -> Callable[[], Settings]:
    settings = Settings()

    def get_settings() -> Settings:
        return settings

    return get_settings


get_settings = _configure_initial_settings()


from .application import Application
from .measurement import Measurement
from .default_logger import DefaultLogger
from .logger import LoggerSingleton

__all__ = [
    "Application",
    "Measurement",
    "DefaultLogger",
    "LoggerSingleton"
]


from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from ..logger import LoggerSingleton
from datetime import datetime
from ..logs.measurement import Measurement
from ..logs.application import Application
from pvmlib.schemas.errors_schema import ErrorGeneralSchema, ErrorDataSchema, ErrorMetaSchema
from fastapi.responses import JSONResponse




class InterceptMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, app_info: Application, auth_service_url: str = None):
        super().__init__(app)
        self.app_info = app_info
        self.logger = LoggerSingleton().get_logger()
        self.auth_service_url = auth_service_url

    async def dispatch(self, request: Request, call_next):
        if "healthcheck" in request.url.path:
            return await call_next(request)

        start_time = datetime.now()
        
        measurement = Measurement(
            service=self.app_info.name,
            method=request.method,
            message="Success",
            time_elapsed=0
        )

        response = None
        
        try:
            response = await call_next(request)
            elapsed_time = (datetime.now() - start_time).total_seconds() * 1000
            measurement.time_elapsed = int(elapsed_time)
            measurement.message = "success"
            self.logger.info(f"Request processed successfully: {request.url.path}", extra={"measurement": measurement})
            return response
        except Exception as e:
            elapsed_time = (datetime.now() - start_time).total_seconds() * 1000
            measurement.time_elapsed = int(elapsed_time)
            measurement.message = "error"
            self.logger.error(f"Request failed: {request.url.path}", exc_info={str(e)}, extra={"measurement": measurement})
            
            error_response = ErrorGeneralSchema(
                error=ErrorDataSchema(user_message="An unexpected error occurred."),
                trace=ErrorMetaSchema(
                    error_code=500,
                    info=str(e)
                )
            )
            
            return JSONResponse(content=error_response.model_dump(), status_code=500)


class LibraryInfo:
    version_lib = 'v1.6.5'
    name = 'pvmlib'
    author = "Jesus Lizarraga"
    author_email = "jesus.lizarragav@coppel.com"
    description = "Python library for PVM"
    python_requires = '>=3.12'
    env = 'development'
    
    install_requires = [
        "pydantic-settings",
        "pydantic",
        "fastapi",
        "uvicorn",
        "pytz",
        "circuitbreaker",
        "requests",
        "tenacity",
        "pybreaker",
        "motor",
        "starlette>=0.13.0",
        "urllib3>=1.26.5,<2.0.0",
        "charset_normalizer>=2.0.0,<3.0.0"
    ]


from pydantic import BaseModel, Field

class DataSchema(BaseModel):
    """
        Data Schema
    """
    status: str = Field(
        title="Status",
        description="Estatus de los servicios utilizados"
    )


class ErrorDataSchema(BaseModel):
    """
        Error Data Schema
    """
    user_message: str = Field(
        title="User Message",
        description="Mensaje de error para el usuario"
    )


class MetaSchema(BaseModel):
    """
        Meta Schema
    """
    time_elapsed: float = Field(
        default=1.0,
        title="Time elapsed",
        description="Tiempo que duró la operación"
    )
    timestamp: str = Field(
        default="2023-02-24T09:09:39.196026",
        title="Timestamp",
        description="Fecha en formato UTC"
    )
    transaction_id: str = Field(
        default="b284993c-323b-455a-87b9-ebf70b62bfa2",
        title="Transaction ID",
        description="Identificador único de la operación"
    )


class ErrorMetaSchema(BaseModel):
    """
        Error Meta Schema
    """
    details: list | dict | str | None = Field(
        title="Error Details",
        description="Datos adicionales del error",
        default=None
    )
    error_code: int = Field(
        title="Error Code",
        description="Identificador único del error"
    )
    info: str = Field(
        title="Info",
        description="URL donde se encuentra información más detallada del error"
    )
    timestamp: str = Field(
        default="2023-02-24T09:09:39.196026",
        title="Timestamp",
        description="Fecha en formato UTC"
    )
    transaction_id: str = Field(
        default="b284993c-323b-455a-87b9-ebf70b62bfa2",
        title="Transaction ID",
        description="Identificador único de la operación"
    )


from pydantic import BaseModel
from typing import Any, Dict

#Schema general de respuesta
class ResponseMetaSchema(BaseModel):
    id_transaction: str
    status: str
    time_elapsed: float

class ResponseGeneralSchema(BaseModel):
    code: int = 0
    type: int = 0
    message: str = "SUCCESS"
    data: Dict[str, Any]
    meta: ResponseMetaSchema

type_content = "application/json"

success_general_schema = {
    "definitions": {
        "ResponseGeneralSchema": {
            "type": "object",
            "properties": {
                "code": {"type": "integer"},
                "type": {"type": "integer"},
                "message": {"type": "string"},
                "data": {"type": "object"},
                "meta": {
                    "type": "object",
                    "properties": {
                        "id_transaction": {"type": "string"},
                        "status": {"type": "string"},
                        "time_elapsed": {"type": "number"}
                    }
                }
            }
        }
    }
}

success_response_example = {
    "description": "Successful Response",
    "content": {
        type_content: {
            "schema": success_general_schema["definitions"]["ResponseGeneralSchema"],
            "examples": {
                "SuccessExample": {
                    "value": ResponseGeneralSchema(
                        code=0,
                        type=0,
                        message="Success",
                        data={"dns": "https://pvm-stg.coppel.io/puntodeventa/api/v1/"},
                        meta=ResponseMetaSchema(
                            id_transaction="bdae4708-72fd-49a8-8f73-646605101072",
                            status="SUCCESS",
                            time_elapsed=2.7139980792999268
                        )
                    ).model_dump()
                }
            }
        }
    }
}


from abc import ABCMeta, abstractmethod


class MongoRepository(metaclass=ABCMeta):
    @abstractmethod
    def is_up(self, log) -> bool:
        raise NotImplementedError
    
    
    @abstractmethod
    def config_general(self, log, *args, **kwargs) -> dict:
        raise NotImplementedError


from contextlib import contextmanager
import abc
import motor.motor_asyncio
from pvmlib.logger import LoggerSingleton
from tenacity import retry, stop_after_attempt, wait_fixed

logger = LoggerSingleton().get_logger()

class Session:
    pass

class Database:
    @abc.abstractmethod
    @contextmanager
    def session(self): pass

class DatabaseManager(Database):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(DatabaseManager, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        if not hasattr(self, 'initialized'):
            self.mongo_database = None
            self.mongo_client = None
            self.initialized = True

    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2))
    async def connect_to_mongo(self, settings_env):
        try:
            self.mongo_client = motor.motor_asyncio.AsyncIOMotorClient(settings_env.MONGO_URI)
            self.mongo_database = self.mongo_client[settings_env.MONGO_DB_NAME]
        except Exception as e:
            logger.error(f"Failed to connect to MongoDB: {e}")
            raise

    async def disconnect_from_mongo(self):
        self.mongo_client.close()

    @contextmanager
    def session(self):
        try:
            yield self.mongo_database
        finally:
            pass 

database_manager = DatabaseManager()


class Measurement(object):

    def __init__(self, service: str, method: str, time_elapsed: int | float, message: str = "Success") -> None:
        self.service = service
        self.method = method
        self.message = message
        self.time_elapsed = time_elapsed

    def get_service(self) -> dict:
        return {
            "service": self.service,
            "method": self.method,
            "message": self.message,
            "time_elapsed": self.time_elapsed
        }


class Application(object):

    def __init__(self, name: str, version: str, env: str, kind: str) -> None:
        self.name = name
        self.version = version
        self.env = env
        self.kind = kind

    def get_info_application(self) -> dict:
        return {
            "name": self.name,
            "version": self.version,
            "env": self.env,
            "kind": self.kind
        }


from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from .response_config.response import Response


class HttpResponse(JSONResponse):

    def __init__(self, status_code: int, content: Response, excludes: set = {}, *args, **kwargs) -> None:
        content_data = jsonable_encoder(content, exclude=excludes)
        super().__init__(
            content=content_data,
            status_code=status_code,
            *args,
            **kwargs
        )



from pydantic import BaseModel
from typing import Any, Dict, Optional

#Schema general de respuesta de error
class ErrorDataSchema(BaseModel):
    user_message: str

class ErrorMetaSchema(BaseModel):
    error_code: int
    info: str

class ErrorGeneralSchema(BaseModel):
    error: ErrorDataSchema
    trace: ErrorMetaSchema

type_content = "application/json"

error_general_schema = {
    "definitions": {
        "ErrorGeneralSchema": {
            "type": "object",
            "properties": {
                "error": {
                    "type": "object",
                    "properties": {
                        "user_message": {"type": "string"}
                    }
                },
                "trace": {
                    "type": "object",
                    "properties": {
                        "error_code": {"type": "integer"},
                        "info": {"type": "string"}
                    }
                }
            }
        }
    }
}

error_response_general_405 = {
    "description": "Method Not Allowed",
    "content": {
        type_content: {
            "schema": error_general_schema["definitions"]["ErrorGeneralSchema"],
            "examples": {
                "MethodNotAllowed": {
                    "value": ErrorGeneralSchema(
                        error=ErrorDataSchema(user_message="Método no permitido."),
                        trace=ErrorMetaSchema(
                            error_code=405,
                            info="405"
                        )
                    ).model_dump()
                }
            }
        }
    }
}


class Utils:
    @staticmethod
    def get_method_name(obj, func_name: str = '') -> str:
        obj_class_name = f"{obj.__class__.__module__}.{obj.__class__.__qualname__}"
        full_name = obj_class_name + '.' + func_name if func_name else obj_class_name
        return full_name
    

    @staticmethod
    def add_attributes(obj, data: dict) -> None:
        for key, value in data.items():
            setattr(obj, key, value)
    

    @staticmethod
    def discard_empty_attributes(obj) -> None:
        obj_copy = obj.__dict__.copy()
        for key, value in obj_copy.items():
            if not value:
                delattr(obj, key)
    

    @staticmethod
    def sort_attributes(obj) -> None:
        obj.__dict__ = dict(sorted(obj.__dict__.items()))

    
    @staticmethod
    def get_error_data(error_details: dict, error_code: str) -> dict:
        return {
            "message": error_details[error_code]["message"],
            "status_code": int(error_details[error_code]["code_name"].split(".")[0]),
            "code_name": error_details[error_code]["code_name"]
        }
    

    @staticmethod
    def get_error_details(errors: dict) -> list:
        return list(map(lambda error: f"{error['loc'][1]}: {error['msg']} in {error['loc'][0]}", errors))
    

    @staticmethod
    def get_error_details_query_params(errors: dict) -> list:
        details = list(map(lambda error: f"{error['loc'][0]}: {error['msg']} in {error['loc'][1]}", errors))
        if "__root__" in details[0]:
            detail = details[0].replace("__root__", "puntodeventa")
            details[0] = detail
        return details



from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware
from fastapi.responses import JSONResponse
from pvmlib.schemas.success_schema import ResponseGeneralSchema, ResponseMetaSchema
from pvmlib.schemas.errors_schema import ErrorGeneralSchema, ErrorDataSchema, ErrorMetaSchema
from pvmlib.code_diccionary import HTTP_STATUS_CODE
import json
from datetime import datetime
import uuid

class AddDataMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        if "healthcheck" in request.url.path:
            return await call_next(request)
        
        start_time = datetime.now()
        
        response = await call_next(request)
        time_elapsed = (datetime.now() - start_time).total_seconds() * 1000
        response_body = [section async for section in response.__dict__['body_iterator']]
        response_body = json.loads(response_body[0].decode())

        # Remove type, message, and code from response_body if they exist
        response_body.pop("type", None)
        response_body.pop("message", None)
        response_body.pop("code", None)

        status_message = HTTP_STATUS_CODE.get(response.status_code, "Unknown Status Code")
        new_response_body = ResponseGeneralSchema(
            type=response_body.get("type", 0),
            message=response_body.get("message", "OK"),
            code=response_body.get("code", 0),
            data=response_body,
            meta=ResponseMetaSchema(
                id_transaction=str(uuid.uuid4()),  # Convert UUID to string
                status=status_message,
                time_elapsed=time_elapsed
            )
        )

        return JSONResponse(content=new_response_body.model_dump(), status_code=response.status_code)


from .application import Application
from .measurement import Measurement
from typing import Dict, Any

class DefaultLogger(object):
 
    def __init__(
            self, 
            level: str, 
            schemaVersion: str,
            logType: str,
            sourceIP: str,
            status: str,
            message: str,
            logOrigin: str,
            timestamp: str,
            tracingId: str,
            hostname: str,
            eventType: str,
            application: Application,
            measurement: Measurement,
            destinationIP: str,
            additionalInfo: Dict[str, Any]) -> None:

        self.level = level
        self.schemaVersion = schemaVersion
        self.logType = logType
        self.sourceIP = sourceIP
        self.status = status
        self.message = message
        self.logOrigin = logOrigin
        self.timestamp = timestamp
        self.tracingId = tracingId
        self.hostname = hostname
        self.eventType = eventType
        self.application = application
        self.measurement = measurement
        self.destinationIP = destinationIP
        self.additionalInfo = additionalInfo
    
    def get_default_log(self) -> dict:
        return {
           "level": self.level,
           "schemaVersion": self.schemaVersion,
           "logType": self.logType,
           "sourceIP": self.sourceIP,
           "status": self.status,
           "message": self.message,
           "logOrigin": self.logOrigin,
           "timestamp": self.timestamp,
           "tracingId": self.tracingId,
           "hostname": self.hostname,
           "eventType": self.eventType,
           "application": self.application.get_info_application(),
           "measurement": self.measurement.get_service(),
           "destinationIP": self.destinationIP, 
           "additionalInfo": self.additionalInfo
        }


from contextlib import asynccontextmanager
from fastapi import FastAPI
from .database import database_manager
from .settings import get_settings
from .logs.logger import LoggerSingleton

logger = LoggerSingleton().get_logger()
setting_env = get_settings()

@asynccontextmanager
async def lifespan(app: FastAPI):
        try:
            if "MONGO_URI" in setting_env.model_fields:
                await database_manager.connect_to_mongo(settings_env=setting_env)
                logger.info("Connected to MongoDB")
        except Exception as e:
            logger.error(f"Error in connect to MongoDB: {e}")
            raise e

        yield
        # Shutdown
        if "MONGO_URI" in setting_env:
            logger.info("Disconnecting from MongoDB")
            await database_manager.disconnect_from_mongo()


HTTP_STATUS_CODE = {
    200: 'SUCCESS',
    201: 'CREATE',
    204: 'NO CONTENT',
    400: 'BAD REQUEST',
    401: 'UNAUTHORIZED',
    403: 'FORBIDDEN',
    404: 'NOT FOUND',
    405: 'METHOD NOT ALLOWED',
    409: 'CONFLICT',
    500: 'INTERNAL SERVER ERROR',
    422: 'UNPROCESSABLE ENTITY',
    429: 'TOO MANY REQUESTS',
}


from pydantic import BaseModel, Field
from .general_schema import (
    DataSchema, 
    MetaSchema
)
from .errors_schema import (
    error_response_general_405,
)


class ReadinessSchema(BaseModel):
    """
        Readiness Schema
    """
    data: DataSchema = Field(
        default=DataSchema(status="Mongo are up"),
        title="DataSchema"
    )
    meta: MetaSchema = Field(...)


responses_readiness = {
    200: {"model": ReadinessSchema},
    405: error_response_general_405
    }


import re
import json

def mask_half_sensitive_info(match):
    match_str = match.group(0)
    half_length = len(match_str) // 2
    return match_str[:half_length] + '****'

def mask_sensitive_info(log_message):
    sensitive_patterns = [
        r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b',
        r'\b\d{3}[-.\s]??\d{2}[-.\s]??\d{4}\b',
        r'\b\d{4}[-.\s]??\d{4}[-.\s]??\d{4}[-.\s]??\d{4}\b',
        r'\b\d{10}\b',
        r'\b\d{6,10}\b',
        r'\b[A-Z][a-z]*\b',
        r'\b[A-Z][a-z]*\b',
    ]
    
    for pattern in sensitive_patterns:
        log_message = re.sub(pattern, mask_half_sensitive_info, log_message)
    
    return log_message

def mask_sensitive_info_in_json(log_json):
    try:
        log_data = json.loads(log_json)
        masked_log_data = {key: mask_sensitive_info(str(value)) for key, value in log_data.items()}
        return json.dumps(masked_log_data)
    except json.JSONDecodeError:
        return log_json

def sensitive_info_decorator(func):
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, str):
            try:
                json.loads(result)
                result = mask_sensitive_info_in_json(result)
            except json.JSONDecodeError:
                result = mask_sensitive_info(result)
        return result
    return wrapper



from motor.motor_asyncio import AsyncIOMotorClient
from ..logs import Measurement
from .mongo_repository import MongoRepository
from datetime import datetime
import time

class MongoWorkerRepository(MongoRepository):

    get_time_elapsed_ms = lambda self, init_time=None, decimals=2: round((time.perf_counter()-init_time)*1000, decimals)

    def __init__(self, database_url: str) -> None:
        self.client = AsyncIOMotorClient(database_url)
        self.database = self.client.get_default_database()

    async def is_up(self, log) -> bool:
        init_time = time.perf_counter()
        try:
            await self.database.command("ping")
            time_elapsed = self.get_time_elapsed_ms(init_time)
            measurement = Measurement("MongoDB", time_elapsed)
            log.info("MongoDB is up", measurement=measurement)
            return True
        except Exception as e:
            time_elapsed = self.get_time_elapsed_ms(init_time)
            measurement = Measurement("MongoDB", time_elapsed, "error")
            log.error("MongoDB is not up", error=str(e), measurement=measurement)
            return False



from fastapi import APIRouter, HTTPException
from pvmlib.settings import get_settings
from pvmlib.schemas.readiness_schema import responses_readiness
from pvmlib.mongo_worker.mongo_worker_repository import MongoWorkerRepository
from pvmlib.logger import LoggerSingleton

settings = get_settings()
readiness_router = APIRouter()
logger = LoggerSingleton().get_logger()

@readiness_router.get(
    "/healthcheck/readness", 
    tags=["Health Check"], 
    responses=responses_readiness, 
    summary="Status de los recursos"
)
async def readiness() -> dict:
    """ Comprueba que el servicio este operando """
    try:
        database_url = settings.MONGO_URI
        if not database_url:
            logger.error("URL_DB environment variable not found")
            raise Exception("URL_DB environment variable not found")
        # Aquí puedes agregar más lógica para comprobar la disponibilidad de otros recursos
        return {"status": "UP"}
    except Exception as e:
        logger.error(f"Readiness check failed: {str(e)}")
        raise HTTPException(status_code=500, detail="Readiness check failed")


from pvmlib.logs import application, measurement, default_logger, logger
from pvmlib.database import DatabaseManager as database_manager
from pvmlib.middleware.request_middleware import InterceptMiddleware as intercept_middleware
from pvmlib.middleware.data_response_middleware import AddDataMiddleware as data_response_middleware
from pvmlib.circuit_breaker import circuit_breaker
from pvmlib.decorator import sensitive_info_decorator
from pvmlib.mongo_worker import MongoRepository, MongoWorkerRepository
from pvmlib.healthcheck.liveness import liveness_router
from pvmlib.healthcheck.readiness import readiness_router
from pvmlib.utils import Utils
from pvmlib.settings import Settings
from pvmlib.http_response import HttpResponse
from pvmlib.cache_config import Cache
from pvmlib.info_lib import LibraryInfo
from pvmlib.response_config.error_response import ErrorResponse, RequestValidationError
from pvmlib.response_config.error_response import (
     internal_server_error_exception_handler, 
     error_exception_handler,
     not_found_error_exception_handler,
     parameter_exception_handler,
     method_not_allowed_exception_handler,
     error_exception
)
from .lifespan import lifespan


name = 'pvmlib'

__all__ = [
    "logger",
    "application",
    "default_logger",
    "measurement",
    "database_manager",
    "liveness_router",
    "readiness_router",
    "intercept_middleware",
    "data_response_middleware",
    "circuit_breaker",
    "sensitive_info_decorator",
    "MongoRepository",
    "MongoWorkerRepository",
    "Utils",
    "Settings",
    "HttpResponse",
    "Cache",
    "LibraryInfo",
    "RequestValidationError",
    "internal_server_error_exception_handler", 
    "error_exception_handler",
    "not_found_error_exception_handler",
    "parameter_exception_handler",
    "method_not_allowed_exception_handler",
    "internal_server_error_exception_handler",
    "error_exception",
    "lifespan"
]


from fastapi import APIRouter, Depends
from pvmlib.settings import get_settings 
from pvmlib.schemas.liveness_schema import (
    responses_liveness,
)

settings = get_settings()

liveness_router = APIRouter()


@liveness_router.get(
    "/healthcheck/liveness", 
    tags=["Health Check"], 
    responses=responses_liveness, 
    summary="Status del servicio"
)
def liveness() -> dict:
    """ Comprueba que el servicio este operando """
    return {"status": "UP"}



from setuptools import setup, find_packages
from pvmlib.info_lib import LibraryInfo

setup(
    name=LibraryInfo.name,
    version=LibraryInfo.version_lib.lstrip("v"),
    packages=find_packages(),
    install_requires=LibraryInfo.install_requires,
    author=LibraryInfo.author,
    author_email=LibraryInfo.author_email,
    description=LibraryInfo.description,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    python_requires=LibraryInfo.python_requires,
)


from fastapi import Request, HTTPException
from fastapi.responses import JSONResponse
from fastapi.exceptions import RequestValidationError
from ..utils import Utils
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY, HTTP_405_METHOD_NOT_ALLOWED, HTTP_500_INTERNAL_SERVER_ERROR, HTTP_400_BAD_REQUEST
from pvmlib.schemas.errors_schema import ErrorGeneralSchema, ErrorDataSchema, ErrorMetaSchema
from fastapi.responses import JSONResponse

class ErrorResponse(Exception):
    def __init__(
            self, error_code: str, message: str, transaction_id: str,
            status_code: int = 500, info: str = None,
            reference_code: str = None, **kwargs
        ) -> None:
        self._status_code = status_code
        self.data = {
            "user_message": message
        }
        self.meta = {
            "error_code": int(error_code),
            "transaction_id": transaction_id,
            "info": info,
            "reference_code": reference_code,
            **kwargs
        }

async def error_exception(exc: Exception):
    error_response = ErrorGeneralSchema(
        error=ErrorDataSchema(user_message="An unexpected error occurred."),
        trace=ErrorMetaSchema(
            error_code=500,
            info=str(exc.__traceback__)
        )
    )
    return JSONResponse(content=error_response.model_dump(), status_code=500)


async def internal_server_error_exception_handler(request: Request, exc: HTTPException):
    error_response = ErrorGeneralSchema(
        error=ErrorDataSchema(user_message="An unexpected error occurred."),
        trace=ErrorMetaSchema(
            error_code=HTTP_500_INTERNAL_SERVER_ERROR,
            info=str(exc.detail) if exc.detail else "Internal Server Error"
        )
    )
    return JSONResponse(content=error_response.model_dump(), status_code=HTTP_500_INTERNAL_SERVER_ERROR)

async def error_exception_handler(request: Request, exc: HTTPException):
    status_code = getattr(exc, 'status_code', HTTP_500_INTERNAL_SERVER_ERROR)
    detail = getattr(exc, 'detail', "An unexpected error occurred.")
    error_response = ErrorGeneralSchema(
        error=ErrorDataSchema(user_message=detail),
        trace=ErrorMetaSchema(
            error_code=status_code,
            info=str(detail)
        )
    )
    return JSONResponse(content=error_response.model_dump(), status_code=status_code)

async def not_found_error_exception_handler(request: Request, exc: HTTPException):
    error_response = ErrorGeneralSchema(
        error=ErrorDataSchema(user_message="Resource not found."),
        trace=ErrorMetaSchema(
            error_code=exc.status_code,
            info=exc.detail
        )
    )
    return JSONResponse(content=error_response.model_dump(), status_code=exc.status_code)

async def parameter_exception_handler(request: Request, exc: RequestValidationError):
    error_details = Utils.get_error_details(exc.errors())
    error_response = ErrorGeneralSchema(
        error=ErrorDataSchema(user_message="Validation error."),
        trace=ErrorMetaSchema(
            error_code=HTTP_422_UNPROCESSABLE_ENTITY,
            info="; ".join(error_details)
        )
    )
    return JSONResponse(content=error_response.model_dump(), status_code=HTTP_422_UNPROCESSABLE_ENTITY)

async def method_not_allowed_exception_handler(request: Request, exc: HTTPException):
    error_response = ErrorGeneralSchema(
        error=ErrorDataSchema(user_message="Method not allowed."),
        trace=ErrorMetaSchema(
            error_code=HTTP_405_METHOD_NOT_ALLOWED,
            info=str(exc.detail) if exc.detail else "Method Not Allowed"
        )
    )
    return JSONResponse(content=error_response.model_dump(), status_code=HTTP_405_METHOD_NOT_ALLOWED)

async def bad_request_exception_handler(request: Request, exc: HTTPException):
    error_response = ErrorGeneralSchema(
        error=ErrorDataSchema(user_message="Bad request."),
        trace=ErrorMetaSchema(
            error_code=HTTP_400_BAD_REQUEST,
            info=exc.detail
        )
    )
    return JSONResponse(content=error_response.model_dump(), status_code=HTTP_400_BAD_REQUEST)


from pvmlib.database import Database, Session
from pymongo import MongoClient, errors
from contextlib import contextmanager, suppress
from collections.abc import Iterator
from pvmlib.utils import Utils

class MongoSession(Session):
    def __init__(self, db_uri: str, database: str, timeout: int, max_pool_size: int) -> None:
        self.__client = MongoClient(
            db_uri, 
            serverSelectionTimeoutMS=timeout,
            maxPoolSize=max_pool_size
        )
        self.__db_appcom = self.__client[database]

    def __enter__(self): 
        return self

    def db_appcom(self):
        return self.__db_appcom

    def is_up(self) -> dict:
        data = {
            "status": True,
            "message": "Success",
            "method": Utils.get_method_name(self, "is_up")
        }

        try:
            self.__client.server_info()
            self.__client.admin.command("ping")
        except errors.PyMongoError as ex:
            data["status"] = False
            data["message"] = str(ex)
        
        return data

    def __exit__(self, exception_type, exception_value, traceback) -> None: 
        self.__client.close()


class MongoDatabase(Database):
    def __init__(self, db_uri: str, database: str, timeout: int, max_pool_size: int) -> None:
        self.__session = MongoSession(
            db_uri,
            database,
            timeout,
            max_pool_size
        )

    @contextmanager
    def session(self) -> Iterator[Session]:
        with suppress(Exception):
            yield self.__session


from .logs import Measurement
import time


class Cache(object):

    get_time_elapsed_ms = lambda self, init_time=None, decimals=2: round((time.perf_counter()-init_time)*1000, decimals)
    cached_responses = {}

    def __call__(self, func):
        def get_config_cached(*args, **kwargs):
            full_name = func.__qualname__
            log = kwargs.get("log")
            update_cache = kwargs.get("update_cache")
            is_cached = self.cached_responses.get(full_name) != None
            if is_cached:
                is_cached = self.cached_responses.get(full_name)["data"] != None
            if not is_cached or update_cache:
                res = func(*args, **kwargs)
                self.cached_responses[full_name] = {"data": res}
            elif log: 
                init_time = time.perf_counter()
                measurement = Measurement("MongoDB", self.get_time_elapsed_ms(init_time))
                log.info(f"Get cache data from MongoDB: {full_name}", measurement=measurement)
            return self.cached_responses[full_name]["data"]
        return get_config_cached
    
    @classmethod
    def get_cached_response(cls, func_name: str) -> dict:
        return cls.cached_responses.get(func_name)


from pydantic import BaseModel, Field
from .errors_schema import (
    error_response_general_405
)


class LivenessSchema(BaseModel):
    """
        Liveness Schema
    """
    status: str = Field(
        default="success",
        title="Status",
        description="Estatus del servicio"
    )


responses_liveness = {
    200: {"model": LivenessSchema},
    405: error_response_general_405
}


from pvmlib.utils import Utils
from datetime import datetime


class MetadataResponse:

    transaction_id: str
    timestamp: datetime.now
    time_elapsed: int | float = None
    
    def __init__(self, transaction_id: str, time_elapsed: int | float = None, **kwargs) -> None:
        self.transaction_id = transaction_id
        self.timestamp = datetime.now()
        self.time_elapsed = time_elapsed
        Utils.add_attributes(self, kwargs)
        Utils.discard_empty_attributes(self)
        Utils.sort_attributes(self)


class Response:

    data: dict
    meta: MetadataResponse

    def __init__(self, data: dict, meta: MetadataResponse) -> None:
        self.data = data
        self.meta = meta    


class SuccessResponse(Response):
    def __init__(self, data: dict, status_code: int, transaction_id: str, time_elapsed: int | float, **kwargs) -> None:
        self._status_code = status_code
        meta = MetadataResponse(transaction_id, time_elapsed, **kwargs)
        super().__init__(data, meta)


class FailureResponse(Response):
    def __init__(self, data: dict, status_code: int, transaction_id: str, **kwargs) -> None:
        self._status_code = status_code
        meta = MetadataResponse(transaction_id, **kwargs)
        super().__init__(data, meta)